@@ -45,17 +45,56 @@
 #include "util.h"
 #include "template.h"
 #include "base64.h"
+#include "url-parser/url_parser.h"
 
 using namespace nghttp2;
 
 namespace shrpx {
 
+namespace {
+int htp_msg_begin(llhttp_t *htp);
+int htp_uricb(llhttp_t *htp, const char *data, size_t len);
+int htp_hdr_keycb(llhttp_t *htp, const char *data, size_t len);
+int htp_hdr_valcb(llhttp_t *htp, const char *data, size_t len);
+int htp_hdrs_completecb(llhttp_t *htp);
+int htp_bodycb(llhttp_t *htp, const char *data, size_t len);
+int htp_msg_completecb(llhttp_t *htp);
+} // namespace
+
+namespace {
+constexpr llhttp_settings_t htp_hooks = {
+    htp_msg_begin,       // llhttp_cb      on_message_begin;
+    htp_uricb,           // llhttp_data_cb on_url;
+    nullptr,             // llhttp_data_cb on_status;
+    nullptr,             // llhttp_data_cb on_method;
+    nullptr,             // llhttp_data_cb on_version;
+    htp_hdr_keycb,       // llhttp_data_cb on_header_field;
+    htp_hdr_valcb,       // llhttp_data_cb on_header_value;
+    nullptr,             // llhttp_data_cb on_chunk_extension_name;
+    nullptr,             // llhttp_data_cb on_chunk_extension_value;
+    htp_hdrs_completecb, // llhttp_cb      on_headers_complete;
+    htp_bodycb,          // llhttp_data_cb on_body;
+    htp_msg_completecb,  // llhttp_cb      on_message_complete;
+    nullptr,             // llhttp_cb      on_url_complete;
+    nullptr,             // llhttp_cb      on_status_complete;
+    nullptr,             // llhttp_cb      on_method_complete;
+    nullptr,             // llhttp_cb      on_version_complete;
+    nullptr,             // llhttp_cb      on_header_field_complete;
+    nullptr,             // llhttp_cb      on_header_value_complete;
+    nullptr,             // llhttp_cb      on_chunk_extension_name_complete;
+    nullptr,             // llhttp_cb      on_chunk_extension_value_complete;
+    nullptr,             // llhttp_cb      on_chunk_header;
+    nullptr,             // llhttp_cb      on_chunk_complete;
+    nullptr,             // llhttp_cb      on_reset;
+};
+} // namespace
+
 HttpsUpstream::HttpsUpstream(ClientHandler *handler)
     : handler_(handler),
       current_header_length_(0),
       ioctrl_(handler->get_rlimit()),
       num_requests_(0) {
-  http_parser_init(&htp_, HTTP_REQUEST);
+  llhttp_init(&htp_, HTTP_REQUEST, &htp_hooks);
   htp_.data = this;
 }
 
@@ -87,7 +126,7 @@ void HttpsUpstream::on_start_request() {
 }
 
 namespace {
-int htp_msg_begin(http_parser *htp) {
+int htp_msg_begin(llhttp_t *htp) {
   auto upstream = static_cast<HttpsUpstream *>(htp->data);
   upstream->on_start_request();
   return 0;
@@ -95,7 +134,7 @@ int htp_msg_begin(http_parser *htp) {
 } // namespace
 
 namespace {
-int htp_uricb(http_parser *htp, const char *data, size_t len) {
+int htp_uricb(llhttp_t *htp, const char *data, size_t len) {
   auto upstream = static_cast<HttpsUpstream *>(htp->data);
   auto downstream = upstream->get_downstream();
   auto &req = downstream->request();
@@ -114,7 +153,8 @@ int htp_uricb(http_parser *htp, const char *data, size_t len) {
     assert(downstream->get_request_state() == DownstreamState::INITIAL);
     downstream->set_request_state(
         DownstreamState::HTTP1_REQUEST_HEADER_TOO_LARGE);
-    return -1;
+    llhttp_set_error_reason(htp, "too long request URI");
+    return HPE_USER;
   }
 
   req.fs.add_extra_buffer_size(len);
@@ -131,7 +171,7 @@ int htp_uricb(http_parser *htp, const char *data, size_t len) {
 } // namespace
 
 namespace {
-int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
+int htp_hdr_keycb(llhttp_t *htp, const char *data, size_t len) {
   auto upstream = static_cast<HttpsUpstream *>(htp->data);
   auto downstream = upstream->get_downstream();
   auto &req = downstream->request();
@@ -146,7 +186,8 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
       downstream->set_request_state(
           DownstreamState::HTTP1_REQUEST_HEADER_TOO_LARGE);
     }
-    return -1;
+    llhttp_set_error_reason(htp, "too large header");
+    return HPE_USER;
   }
   if (downstream->get_request_state() == DownstreamState::INITIAL) {
     if (req.fs.header_key_prev()) {
@@ -159,7 +200,8 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
         }
         downstream->set_request_state(
             DownstreamState::HTTP1_REQUEST_HEADER_TOO_LARGE);
-        return -1;
+        llhttp_set_error_reason(htp, "too many headers");
+        return HPE_USER;
       }
       req.fs.alloc_add_header_name(StringRef{data, len});
     }
@@ -173,7 +215,8 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
           ULOG(INFO, upstream)
               << "Too many header field num=" << req.fs.num_fields() + 1;
         }
-        return -1;
+        llhttp_set_error_reason(htp, "too many headers");
+        return HPE_USER;
       }
       req.fs.alloc_add_trailer_name(StringRef{data, len});
     }
@@ -183,7 +226,7 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
 } // namespace
 
 namespace {
-int htp_hdr_valcb(http_parser *htp, const char *data, size_t len) {
+int htp_hdr_valcb(llhttp_t *htp, const char *data, size_t len) {
   auto upstream = static_cast<HttpsUpstream *>(htp->data);
   auto downstream = upstream->get_downstream();
   auto &req = downstream->request();
@@ -198,7 +241,8 @@ int htp_hdr_valcb(http_parser *htp, const char *data, size_t len) {
       downstream->set_request_state(
           DownstreamState::HTTP1_REQUEST_HEADER_TOO_LARGE);
     }
-    return -1;
+    llhttp_set_error_reason(htp, "too large header");
+    return HPE_USER;
   }
   if (downstream->get_request_state() == DownstreamState::INITIAL) {
     req.fs.append_last_header_value(data, len);
@@ -286,7 +330,7 @@ void rewrite_request_host_path_from_uri(BlockAllocator &balloc, Request &req,
 } // namespace
 
 namespace {
-int htp_hdrs_completecb(http_parser *htp) {
+int htp_hdrs_completecb(llhttp_t *htp) {
   int rv;
   auto upstream = static_cast<HttpsUpstream *>(htp->data);
   if (LOG_ENABLED(INFO)) {
@@ -297,6 +341,11 @@ int htp_hdrs_completecb(http_parser *htp) {
 
   auto downstream = upstream->get_downstream();
   auto &req = downstream->request();
+  auto &balloc = downstream->get_block_allocator();
+
+  for (auto &kv : req.fs.headers()) {
+    kv.value = util::rstrip(balloc, kv.value);
+  }
 
   auto lgconf = log_config();
   lgconf->update_tstamp(std::chrono::system_clock::now());
@@ -305,7 +354,7 @@ int htp_hdrs_completecb(http_parser *htp) {
   req.http_major = htp->http_major;
   req.http_minor = htp->http_minor;
 
-  req.connection_close = !http_should_keep_alive(htp);
+  req.connection_close = !llhttp_should_keep_alive(htp);
 
   handler->stop_read_timer();
 
@@ -318,6 +367,10 @@ int htp_hdrs_completecb(http_parser *htp) {
        << "HTTP/" << req.http_major << "." << req.http_minor << "\n";
 
     for (const auto &kv : req.fs.headers()) {
+      if (kv.name == "authorization") {
+        ss << TTY_HTTP_HD << kv.name << TTY_RST << ": <redacted>\n";
+        continue;
+      }
       ss << TTY_HTTP_HD << kv.name << TTY_RST << ": " << kv.value << "\n";
     }
 
@@ -328,15 +381,11 @@ int htp_hdrs_completecb(http_parser *htp) {
   // transfer-encoding is given.  If transfer-encoding is given, leave
   // req.fs.content_length to -1.
   if (method != HTTP_CONNECT && !req.fs.header(http2::HD_TRANSFER_ENCODING)) {
-    // http-parser returns (uint64_t)-1 if there is no content-length
-    // header field.  If we don't have both transfer-encoding, and
-    // content-length header field, we assume that there is no request
-    // body.
-    if (htp->content_length == std::numeric_limits<uint64_t>::max()) {
-      req.fs.content_length = 0;
-    } else {
-      req.fs.content_length = htp->content_length;
-    }
+    // llhttp sets 0 to htp->content_length if there is no
+    // content-length header field.  If we don't have both
+    // transfer-encoding and content-length header field, we assume
+    // that there is no request body.
+    req.fs.content_length = htp->content_length;
   }
 
   auto host = req.fs.header(http2::HD_HOST);
@@ -366,7 +415,6 @@ int htp_hdrs_completecb(http_parser *htp) {
   downstream->inspect_http1_request();
 
   auto faddr = handler->get_upstream_addr();
-  auto &balloc = downstream->get_block_allocator();
   auto config = get_config();
 
   if (method != HTTP_CONNECT) {
@@ -402,12 +450,18 @@ int htp_hdrs_completecb(http_parser *htp) {
 
   downstream->set_request_state(DownstreamState::HEADER_COMPLETE);
 
+  auto &resp = downstream->response();
+
+  if (config->http.require_http_scheme &&
+      !http::check_http_scheme(req.scheme, handler->get_ssl() != nullptr)) {
+    resp.http_status = 400;
+    return -1;
+  }
+
 #ifdef HAVE_MRUBY
   auto worker = handler->get_worker();
   auto mruby_ctx = worker->get_mruby_context();
 
-  auto &resp = downstream->response();
-
   if (mruby_ctx->run_on_request_proc(downstream) != 0) {
     resp.http_status = 500;
     return -1;
@@ -426,7 +480,9 @@ int htp_hdrs_completecb(http_parser *htp) {
     return 0;
   }
 
+#ifdef HAVE_MRUBY
   DownstreamConnection *dconn_ptr;
+#endif // HAVE_MRUBY
 
   for (;;) {
     auto dconn = handler->get_downstream_connection(rv, downstream);
@@ -436,7 +492,6 @@ int htp_hdrs_completecb(http_parser *htp) {
         upstream->redirect_to_https(downstream);
       }
       downstream->set_request_state(DownstreamState::CONNECT_FAIL);
-
       return -1;
     }
 
@@ -451,7 +506,7 @@ int htp_hdrs_completecb(http_parser *htp) {
 #ifdef HAVE_MRUBY
   const auto &group = dconn_ptr->get_downstream_addr_group();
   if (group) {
-    const auto &dmruby_ctx = group->mruby_ctx;
+    const auto &dmruby_ctx = group->shared_addr->mruby_ctx;
 
     if (dmruby_ctx->run_on_request_proc(downstream) != 0) {
       resp.http_status = 500;
@@ -475,9 +530,7 @@ int htp_hdrs_completecb(http_parser *htp) {
     // and let them decide whether responds with 100 Continue or not.
     // For alternative mode, we have no backend, so just send 100
     // Continue here to make the client happy.
-    auto expect = req.fs.header(http2::HD_EXPECT);
-    if (expect &&
-        util::strieq(expect->value, StringRef::from_lit("100-continue"))) {
+    if (downstream->get_expect_100_continue()) {
       auto output = downstream->get_response_buf();
       constexpr auto res = StringRef::from_lit("HTTP/1.1 100 Continue\r\n\r\n");
       output->append(res);
@@ -490,7 +543,7 @@ int htp_hdrs_completecb(http_parser *htp) {
 } // namespace
 
 namespace {
-int htp_bodycb(http_parser *htp, const char *data, size_t len) {
+int htp_bodycb(llhttp_t *htp, const char *data, size_t len) {
   int rv;
   auto upstream = static_cast<HttpsUpstream *>(htp->data);
   auto downstream = upstream->get_downstream();
@@ -503,14 +556,15 @@ int htp_bodycb(http_parser *htp, const char *data, size_t len) {
       return 0;
     }
 
-    return -1;
+    llhttp_set_error_reason(htp, "could not process request body");
+    return HPE_USER;
   }
   return 0;
 }
 } // namespace
 
 namespace {
-int htp_msg_completecb(http_parser *htp) {
+int htp_msg_completecb(llhttp_t *htp) {
   int rv;
   auto upstream = static_cast<HttpsUpstream *>(htp->data);
   if (LOG_ENABLED(INFO)) {
@@ -518,6 +572,13 @@ int htp_msg_completecb(http_parser *htp) {
   }
   auto handler = upstream->get_client_handler();
   auto downstream = upstream->get_downstream();
+  auto &req = downstream->request();
+  auto &balloc = downstream->get_block_allocator();
+
+  for (auto &kv : req.fs.trailers()) {
+    kv.value = util::rstrip(balloc, kv.value);
+  }
+
   downstream->set_request_state(DownstreamState::MSG_COMPLETE);
   rv = downstream->end_upload_data();
   if (rv != 0) {
@@ -528,9 +589,7 @@ int htp_msg_completecb(http_parser *htp) {
       // next request handling (if we don't close the connection).  We
       // first pause parser here just as we normally do, and call
       // signal_write() to run on_write().
-      http_parser_pause(htp, 1);
-
-      return 0;
+      return HPE_PAUSED;
     }
     return -1;
   }
@@ -544,24 +603,10 @@ int htp_msg_completecb(http_parser *htp) {
   }
 
   // Stop further processing to complete this request
-  http_parser_pause(htp, 1);
-  return 0;
+  return HPE_PAUSED;
 }
 } // namespace
 
-namespace {
-constexpr http_parser_settings htp_hooks = {
-    htp_msg_begin,       // http_cb      on_message_begin;
-    htp_uricb,           // http_data_cb on_url;
-    nullptr,             // http_data_cb on_status;
-    htp_hdr_keycb,       // http_data_cb on_header_field;
-    htp_hdr_valcb,       // http_data_cb on_header_value;
-    htp_hdrs_completecb, // http_cb      on_headers_complete;
-    htp_bodycb,          // http_data_cb on_body;
-    htp_msg_completecb   // http_cb      on_message_complete;
-};
-} // namespace
-
 // on_read() does not consume all available data in input buffer if
 // one http request is fully received.
 int HttpsUpstream::on_read() {
@@ -574,7 +619,7 @@ int HttpsUpstream::on_read() {
   }
 
   // downstream can be nullptr here, because it is initialized in the
-  // callback chain called by http_parser_execute()
+  // callback chain called by llhttp_execute()
   if (downstream && downstream->get_upgraded()) {
 
     auto rv = downstream->push_upload_data_chunk(rb->pos(), rb->rleft());
@@ -609,11 +654,15 @@ int HttpsUpstream::on_read() {
     }
   }
 
-  // http_parser_execute() does nothing once it entered error state.
-  auto nread = http_parser_execute(&htp_, &htp_hooks,
-                                   reinterpret_cast<const char *>(rb->pos()),
-                                   rb->rleft());
+  // llhttp_execute() does nothing once it entered error state.
+  auto htperr = llhttp_execute(&htp_, reinterpret_cast<const char *>(rb->pos()),
+                               rb->rleft());
 
+  auto nread =
+      htperr == HPE_OK
+          ? rb->rleft()
+          : reinterpret_cast<const uint8_t *>(llhttp_get_error_pos(&htp_)) -
+                rb->pos();
   rb->drain(nread);
   rlimit->startw();
 
@@ -624,8 +673,6 @@ int HttpsUpstream::on_read() {
   // execution
   downstream = get_downstream();
 
-  auto htperr = HTTP_PARSER_ERRNO(&htp_);
-
   if (htperr == HPE_PAUSED) {
     // We may pause parser in htp_msg_completecb when both side are
     // completed.  Signal write, so that we can run on_write().
@@ -640,8 +687,8 @@ int HttpsUpstream::on_read() {
   if (htperr != HPE_OK) {
     if (LOG_ENABLED(INFO)) {
       ULOG(INFO, this) << "HTTP parse failure: "
-                       << "(" << http_errno_name(htperr) << ") "
-                       << http_errno_description(htperr);
+                       << "(" << llhttp_errno_name(htperr) << ") "
+                       << llhttp_get_error_reason(&htp_);
     }
 
     if (downstream &&
@@ -732,6 +779,13 @@ int HttpsUpstream::on_write() {
       handler_->repeat_read_timer();
 
       return resume_read(SHRPX_NO_BUFFER, nullptr, 0);
+    } else {
+      // If the request is not complete, close the connection.
+      delete_downstream();
+
+      handler_->set_should_close_after_write(true);
+
+      return 0;
     }
   }
 
@@ -755,7 +809,7 @@ int HttpsUpstream::resume_read(IOCtrlReason reason, Downstream *downstream,
   if (ioctrl_.resume_read(reason)) {
     // Process remaining data in input buffer here because these bytes
     // are not notified by readcb until new data arrive.
-    http_parser_pause(&htp_, 0);
+    llhttp_resume(&htp_);
 
     auto conn = handler_->get_connection();
     ev_feed_event(conn->loop, &conn->rev, EV_READ);
@@ -772,7 +826,10 @@ int HttpsUpstream::downstream_read(DownstreamConnection *dconn) {
   rv = downstream->on_read();
 
   if (rv == SHRPX_ERR_EOF) {
-    return downstream_eof(dconn);
+    if (downstream->get_request_header_sent()) {
+      return downstream_eof(dconn);
+    }
+    return SHRPX_ERR_RETRY;
   }
 
   if (rv == SHRPX_ERR_DCONN_CANCELED) {
@@ -877,7 +934,11 @@ int HttpsUpstream::downstream_error(DownstreamConnection *dconn, int events) {
 
   unsigned int status;
   if (events & Downstream::EVENT_TIMEOUT) {
-    status = 504;
+    if (downstream->get_request_header_sent()) {
+      status = 504;
+    } else {
+      status = 408;
+    }
   } else {
     status = 502;
   }
@@ -1026,18 +1087,6 @@ std::unique_ptr<Downstream> HttpsUpstream::pop_downstream() {
   return std::unique_ptr<Downstream>(downstream_.release());
 }
 
-namespace {
-void write_altsvc(DefaultMemchunks *buf, BlockAllocator &balloc,
-                  const AltSvc &altsvc) {
-  buf->append(util::percent_encode_token(balloc, altsvc.protocol_id));
-  buf->append("=\"");
-  buf->append(util::quote_string(balloc, altsvc.host));
-  buf->append(':');
-  buf->append(altsvc.service);
-  buf->append('"');
-}
-} // namespace
-
 int HttpsUpstream::on_downstream_header_complete(Downstream *downstream) {
   if (LOG_ENABLED(INFO)) {
     if (downstream->get_non_final_response()) {
@@ -1064,7 +1113,7 @@ int HttpsUpstream::on_downstream_header_complete(Downstream *downstream) {
     assert(dconn);
     const auto &group = dconn->get_downstream_addr_group();
     if (group) {
-      const auto &dmruby_ctx = group->mruby_ctx;
+      const auto &dmruby_ctx = group->shared_addr->mruby_ctx;
 
       if (dmruby_ctx->run_on_response_proc(downstream) != 0) {
         error_reply(500);
@@ -1196,13 +1245,7 @@ int HttpsUpstream::on_downstream_header_complete(Downstream *downstream) {
     // We won't change or alter alt-svc from backend for now
     if (!httpconf.altsvcs.empty()) {
       buf->append("Alt-Svc: ");
-
-      auto &altsvcs = httpconf.altsvcs;
-      write_altsvc(buf, downstream->get_block_allocator(), altsvcs[0]);
-      for (size_t i = 1; i < altsvcs.size(); ++i) {
-        buf->append(", ");
-        write_altsvc(buf, downstream->get_block_allocator(), altsvcs[i]);
-      }
+      buf->append(httpconf.altsvc_header_value);
       buf->append("\r\n");
     }
   }
